##
# File:    ActionRegistryIo.py
# Date:    8-April-2010
#
# Updates:
#    7-Sep-2015  jdw -  Add lists to capture order of input's and output's
##
"""
I/O manager for the registry of action definitions.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback
from xml.dom import minidom
from wwpdb.io.misc.FormatOut import FormatOut


class ActionRegistryIo(object):

    """I/O manager class action definitions for the action registry.

     The action registry xml encoding has the following organization::

         <?xml version='1.0' encoding='utf-8'?>
         <actionList version='0.01'>
           <action name="copy">
               <inputList>
                   <wfDataObject name=1>
                       <dataReferenceType>file</dataReferenceType>
                       <contentType>model</contentType>
                       <fileFormat>pdbx</fileFormat>
                       <containerType></containerType>
                       <valueType></valueType>
                       <selectorType></selectorType>
                   </wfDataObject>
               </inputList>
               <outputList>
                   <wfDataObject name=1>
                       <dataReferenceType>file</dataReferenceType>
                       <contentType>model</contentType>
                       <fileFormat>pdbx</fileFormat>
                       <containerType></containerType>
                       <valueType></valueType>
                       <selectorType></selectorType>
                   </wfDataObject>
               </outputList>
               <userParameters>
                   <parameter name="mode">r</parameter>
               </userParameters>
               <internalParameters>
                   <parameter name="force">-f</parameter>
               </internalParameters>
               <moduleName>FileUtils</moduleName>
               <methodName>copyOp</methodName>
           </action>
         </actionList>


    The action registry dictionary is stored in a dictionary using following organization:

    - *actionId*,  a unique action identifier referencing a dictionary with the following keys:

      + INPUT_INFO_LIST & OUTPUT_INFO_LIST,  list of required features of the input and out data objects stored as
        tuples of (reference_type, (data_type,container_type), selector_type)
      + USER_PARAMETER_DICTIONARY,  container for user settable parameters passed to the action method.
      + INTERNAL_PARAMETER_DICTIONARY,  container for internal parameters passed to the action method.
      + MODULE_NAME,   Python module(class) name containing the target method
      + METHOD_NAME,   Python method name

    Method prototype:

    method(inputObjList=[],outputObjList=[],parameterDictionary={})

    """

    def __init__(self, filePath="./examples/resources/actionData.xml", verbose=False, log=sys.stderr):
        #
        self.__lfh = log
        self.__verbose = verbose  # pylint: disable=unused-private-member
        #
        # self.lt = time.strftime("%Y%m%d", time.localtime())
        #
        self.__fileName = filePath
        self.__dict = {}
        self.__setup()
        #

    def __setup(self):
        try:
            self.__dom = minidom.parse(self.__fileName)
            self.__dict = self.__getActionDictionary()
            return True
        except Exception as _e:  # noqa: F841
            self.__lfh.write("+ActionRegistryIo.__setup() - read failed for %s\n" % self.__fileName)
            traceback.print_exc(file=self.__lfh)
            return False

    def getRegistry(self):
        """Returns the dictionary data structure representing the action registry."""
        return self.__dict

    def __getWfDataObjectList(self, el):
        oD = {}
        oL = []
        for child in el.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue
            if len(child.childNodes) == 0:
                continue
            #
            if child.nodeName == "wfDataObject":
                dId = child.getAttributeNode("name").nodeValue
                tD = {"dataReferenceType": None, "contentType": None, "fileFormat": None, "containerType": None, "valueType": None, "selectorType": None}
                for tch in child.childNodes:
                    if tch.nodeName in tD.keys():
                        if len(tch.childNodes) > 0:
                            tD[tch.nodeName] = tch.childNodes[0].nodeValue
                oD[dId] = tD
                oL.append(dId)
        #
        return oD, oL

    def __getActionDictionary(self):
        """Parser for action registry data file.  Builds dictionary of action definitions."""
        aD = {}
        els = self.__dom.getElementsByTagName("action")
        for el in els:
            actionId = el.getAttributeNode("name").nodeValue
            rD = {}
            rD["MODULE_NAME"] = None
            rD["METHOD_NAME"] = None
            rD["INPUT_INFO_LIST"] = {}
            rD["OUTPUT_INFO_LIST"] = {}
            rD["INPUT_NAME_LIST"] = []
            rD["OUTPUT_NAME_LIST"] = []
            rD["USER_PARAMETER_DICT"] = {}
            rD["INTERNAL_PARAMETER_DICT"] = {}
            #
            for child in el.childNodes:
                if child.nodeType != child.ELEMENT_NODE:
                    continue
                if len(child.childNodes) == 0:
                    continue

                if child.nodeName == "moduleName":
                    rD["MODULE_NAME"] = child.childNodes[0].nodeValue

                if child.nodeName == "methodName":
                    rD["METHOD_NAME"] = child.childNodes[0].nodeValue

                elif child.nodeName == "userParameters":
                    pD = {}
                    for tch in child.childNodes:
                        if tch.nodeName == "parameter":
                            ky = tch.getAttributeNode("name").nodeValue
                            if len(tch.childNodes) > 0:
                                val = tch.childNodes[0].nodeValue
                            else:
                                val = None
                            pD[ky] = val
                    rD["USER_PARAMETER_DICT"] = pD

                elif child.nodeName == "internalParameters":
                    pD = {}
                    for tch in child.childNodes:
                        if tch.nodeName == "parameter":
                            ky = tch.getAttributeNode("name").nodeValue
                            if len(tch.childNodes) > 0:
                                val = tch.childNodes[0].nodeValue
                            else:
                                val = None
                            pD[ky] = val
                    rD["INTERNAL_PARAMETER_DICT"] = pD

                elif child.nodeName == "inputList":
                    rD["INPUT_INFO_LIST"], rD["INPUT_NAME_LIST"] = self.__getWfDataObjectList(child)
                elif child.nodeName == "outputList":
                    rD["OUTPUT_INFO_LIST"], rD["OUTPUT_NAME_LIST"] = self.__getWfDataObjectList(child)
                else:
                    pass

            aD[actionId] = rD
        return aD

    def dump(self):
        out = FormatOut()
        out.autoFormat("Action registry file", self.__dict, 3, 3)
        out.writeStream(self.__lfh)



##
# File:    DataSelector.py
# Date:    5-April-2010
#
# Updates:
#
##
"""
Container for data selection criteria.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"


class DataSelector(object):

    """A container for data selection criteria.

    A selector defines the following parameters:

    - targetCategoryName  name of category to which selection is applied
    - targetAttributeNameList [attributeName, attributeName,...] the values of the
      attributes in this list will be returned subject to the selection condition.

    Selection conditions are stored as a list of tuples of name, value and operator:

    - conditionList  [(attributeName,attributeValue,comparisonOp),,...]

    """

    def __init__(self):
        super(DataSelector, self).__init__()
        # self.__verbose = False
        # sys.stderr.write("DataSelector.__init_()\n")
        self.__selectorType = None
        #
        self.__targetCategoryName = None
        self.__targetAttributeList = []
        self.__selectConditionList = []

    def getSelectorType(self):
        """Get the type of selector type.  Currently only attribute value selections are supported."""
        return self.__selectorType

    def setSelectCategoryName(self, categoryName):
        """Set the target category for this selection.

        Returns:

        True for success or False otherwise
        """
        try:
            self.__targetCategoryName = categoryName
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def getSelectCategoryName(self):
        """Get the taget category for this selection."""
        return self.__targetCategoryName

    def addSelectAttributeName(self, attributeName):
        """Add an attribute to the list of attributes for this selection.

        Returns:

        True for success or False otherwise
        """
        try:
            if attributeName not in self.__targetAttributeList:
                self.__targetAttributeList.append(attributeName)
                self.__selectorType = "attribute"
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def getSelectAttributeList(self):
        """Get the list attributes for the current selection"""
        return self.__targetAttributeList

    def addSelectCondition(self, attributeName, attributeValue, comparisonOp="equal"):
        """Add a selection contition to the current selection.

        A selction condition contains the following:
        - attributeName in the target category.
        - attributeValue
        - one of the comparison operators (equals, ...)

        Returns:

        True for success or False otherwise.

        """
        try:
            self.__selectConditionList.append((attributeName, attributeValue, comparisonOp))
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def getSelectConditionList(self):
        """Get the list selection conditions."""
        return self.__selectConditionList



##
# File:    FileUtils.py
# Date:    6-April-2010
#
# Updates:
#
##
"""
Module of file utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import traceback
import shutil
import datetime
import difflib

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase

#


class FileUtils(UtilsBase):

    """Utility class of simple file operations.

    Current supported operations include:
    - copy
    - make directory
    - get size
    - get modification datetime
    - contextual difference

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParmeterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(FileUtils, self).__init__(verbose, log)

    def copyOp(self, **kwargs):
        """Copy operation the file reference from the input object ('src') to the
        file reference in the output object ('dst')
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            iPth = inpObjD["src"].getFilePathReference()
            oPth = outObjD["dst"].getFilePathReference()
            if self._verbose:
                self._lfh.write("+FileUtils.copyOp Input  path %s\n" % iPth)
                self._lfh.write("+FileUtils.copyOp Output path %s\n" % oPth)
            shutil.copyfile(iPth, oPth)
            return True
        except Exception as _e:  # noqa: F841
            if self._verbose:
                traceback.print_exc(file=self._lfh)
            return False

    def copyIfExistsOp(self, **kwargs):
        """Copy operation the file reference from the input object ('src') to the
        file reference in the output object ('dst')

        Test for 'src' file existence  -- always return true -
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            iPth = inpObjD["src"].getFilePathReference()
            oPth = outObjD["dst"].getFilePathReference()
            if self._verbose:
                self._lfh.write("+FileUtils.copyIfExistsOp Input  path %s\n" % iPth)
                self._lfh.write("+FileUtils.copyIfExistsOp Output path %s\n" % oPth)
            if os.access(iPth, os.R_OK):
                shutil.copyfile(iPth, oPth)
            return True
        except Exception as _e:  # noqa: F841
            # if (self._verbose): traceback.print_exc(file=self._lfh)
            return True

    def makeDirOp(self, **kwargs):
        """Create the directory path for the file reference in the input object ('src').
        Set the mode of the directoy according to the value of *mode*.

        Returns:

        True for success or False otherwise.

        """
        try:
            (inpObjD, _outObjD, uD, _pD) = self._getArgs(kwargs)
            newPth = inpObjD["src"].getDirPathReference()
            if not os.path.exists(newPth):
                modeS = uD["mode"]
                modeO = int(modeS, 8)
                os.makedirs(newPth, modeO)
            return True
        except Exception as _e:  # noqa: F841
            if self._verbose:
                traceback.print_exc(file=self._lfh)
            return False

    def sizeOfOp(self, **kwargs):
        """Determine the input file ('src') size in bytes and store this value in the
        output object ('dst')

        Returns:

        True on success or False otherwise
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            newPth = inpObjD["src"].getFilePathReference()
            if os.path.exists(newPth):
                nBytes = os.path.getsize(newPth)
                outObjD["dst"].setValue(nBytes)
                return True
            else:
                outObjD["dst"].setValue(0)
                return False
        except Exception as _e:  # noqa: F841
            if self._verbose:
                traceback.print_exc(file=self._lfh)
            return False

    def mtimeOp(self, **kwargs):
        """Determine the input file ('src') modification time and store this value in the output object ('dst')."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            newPth = inpObjD["src"].getFilePathReference()
            if os.path.exists(newPth):
                tSec = os.path.getmtime(newPth)
                dt = datetime.datetime.fromtimestamp(tSec)
                outObjD["dst"].setValue(dt)
                return True
            else:
                return False
        except Exception as _e:  # noqa: F841
            if self._verbose:
                traceback.print_exc(file=self._lfh)
            return False

    def diffOp(self, **kwargs):
        """Difference the file references from the input objects ('src1' and 'src2')
        and store the line based difference in the output object ('dst')
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            iPth1 = inpObjD["src1"].getFilePathReference()
            iPth2 = inpObjD["src2"].getFilePathReference()
            if self._verbose:
                self._lfh.write("+FileUtils.diffOp Input  path 1 %s\n" % iPth1)
                self._lfh.write("+FileUtils.diffOp Input  path 2 %s\n" % iPth2)

            ifh = open(iPth1, "r")
            aL1 = ifh.readlines()
            ifh.close()

            ifh = open(iPth2, "r")
            aL2 = ifh.readlines()
            ifh.close()
            oL = difflib.context_diff(aL1, aL2, "src1", "src2")
            outObjD["dst"].setValue(oL)
            return True
        except Exception as _e:  # noqa: F841
            if self._verbose:
                traceback.print_exc(file=self._lfh)
            return False



##
# File:    DpUtils.py
# Date:    23-April-2010
#
# Updates:
#
##
"""
Module of data processing utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback
from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility
import os


class DpUtils(UtilsBase):
    """Utility class to perform data processing operations.

    Current supported operations include:
    - polymer linkage distances

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(DpUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """
        #

    def polymerLinkageDistanceOp(self, **kwArgs):
        """Calculate polymer linkage distances -"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            distPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-poly-link-dist")
            dp.exp(distPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+DpUtils.polymerLinkageDistanceOp() - PDBx     file path: %s\n" % pdbxPath)
                self._lfh.write("+DpUtils.polymerLinkageDistanceOp() - Distance file path: %s\n" % distPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def centreOfMassCalculation(self, **kwargs):

        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwargs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "centre-of-mass.log")

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("centre-of-mass")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)

            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+DpUtils.centreOfMassCalculation() - PDBx     file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



##
# File:    ProcessRunner.py
# Date:    1-April-2010
#
# Updates:
# 01-May-2010 jdw add getParameterDict() method and add adjust return values in setParamterDict()
#
##
"""
Classes for accessing  definition of actions/processes, the assignment of data inputs, data outputs,
and process invocation/execution.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback

#
from wwpdb.utils.wf.process.ActionRegistry import ActionRegistry


class ProcessRunner(object):

    """Provide access to action/process definitions, assign data inputs and outputs,
    and manage invocation/execution of processes.

    """

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False
        self.__actionId = None
        self.__userParameterD = {}
        self.__inputD = {}
        self.__outputD = {}
        self.__aReg = ActionRegistry()

    def setInput(self, name, wfDataObject):
        """Set the input data object identified by the input name.

        Returns:

        True on success or false otherwise.
        """
        try:
            self.__inputD[str(name)] = wfDataObject
            return True
        except Exception as _e:  # noqa: F841
            return False

    def setOutput(self, name, wfDataObject):
        """Set the output data object identified by the input name.

        Returns:

        True on success or false otherwise.
        """
        try:
            self.__outputD[str(name)] = wfDataObject
            return True
        except Exception as _e:  # noqa: F841
            return False

    def setAction(self, actionId):
        """Set the name for this action.  The input action identifier should correspond
        to a action defined in the action registry in class `ActionRegistry`.

        Returns:

        True on success or false otherwise.
        """
        self.__actionId = actionId
        if self.__aReg.isDefinedAction(self.__actionId):
            self.__setParameterDictDefault()
            return True
        else:
            return False

    def __setParameterDictDefault(self):
        """Copy the default user adjustable parameters from the action definition
        to a local dictionary.
        """
        self.__userParameterD = {}
        try:
            dd = self.__aReg.getUserParameterDict(self.__actionId)
            for k, v in dd.items():
                self.__userParameterD[k] = v

        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=sys.stderr)
            return False

        return True

    def getParameterDict(self):
        return self.__userParameterD

    def setParameterDict(self, pD):
        """Copy values for input parameters corresponding to any user adjustable
        parameters for the current action.

        Returns:

        True on success or false otherwise.
        """
        self.__userParameterD = {}
        try:
            for k, v in self.__aReg.getUserParameterDict(self.__actionId).items():

                if k in pD:
                    if self.__verbose:
                        self.__lfh.write("+ProcessRunner.setParameterDict() setting parameter %s to %r\n" % (k, pD[k]))
                    self.__userParameterD[k] = pD[k]
                else:
                    if self.__verbose:
                        self.__lfh.write("+ProcessRunner.setParameterDict() using default setting for parameter %s = %r\n" % (k, v))
                    self.__userParameterD[k] = v

            if self.__verbose:
                self.__lfh.write("+ProcessRunner.setParameterDict() parameter settings:\n")
                for k, v in self.getParameterDict().items():
                    self.__lfh.write("+ProcessRunner.setParameterDict() parameter %s = %r\n" % (k, v))
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=sys.stderr)
            return False

        return True

    def preCheck(self):
        """Check that input and output data objects are consistent with the type requirements
        for the specified action.

        Returns:

        True if requirements are satisfied or False otherwise.

        """

        if not self.__aReg.isDefinedAction(self.__actionId):
            self.__lfh.write("preCheck : not a valid action  %s\n" % self.__actionId)
            return False

        if self.__debug:
            self.__lfh.write("preCheck for action %s\n" % self.__actionId)

        numInp = self.__aReg.getInputObjectCount(self.__actionId)
        # number of inputs consistent?
        if numInp != len(self.__inputD):
            if self.__verbose:
                self.__lfh.write("Failed input count\n")
            return False
        #
        # input types consistent?
        if numInp > 0:
            for iN in self.__aReg.getInputObjectNames(self.__actionId):
                if self.__aReg.isSetInputReferenceType(self.__actionId, iN):
                    if self.__aReg.getInputReferenceType(self.__actionId, iN) != self.__inputD[iN].getReferenceType():
                        if self.__verbose:
                            self.__lfh.write("Failed input reference type %s\n" % iN)
                        return False
                if self.__aReg.isSetInputContainerType(self.__actionId, iN):
                    if self.__aReg.getInputContainerType(self.__actionId, iN) != self.__inputD[iN].getContainerTypeName():
                        if self.__verbose:
                            self.__lfh.write("Failed input container type %s\n" % iN)
                        return False
                if self.__aReg.isSetInputValueType(self.__actionId, iN):
                    if self.__aReg.getInputValueType(self.__actionId, iN) != self.__inputD[iN].getValueTypeName():
                        if self.__verbose:
                            self.__lfh.write("Failed input value type %s\n" % iN)
                        return False
                if self.__aReg.isSetInputSelectorType(self.__actionId, iN):
                    if self.__aReg.getInputSelectorType(self.__actionId, iN) != self.__inputD[iN].getSelectorType():
                        if self.__verbose:
                            self.__lfh.write("Failed input selector type %s registered as %s \n" % (iN, self.__aReg.getInputSelectorType(self.__actionId, iN)))
                        return False

                if self.__aReg.isSetInputContentType(self.__actionId, iN):
                    if self.__aReg.getInputContentType(self.__actionId, iN) != self.__inputD[iN].getContentType():
                        if self.__verbose:
                            self.__lfh.write(
                                "Failed input content type %s - %s:%s\n" % (iN, self.__aReg.getInputContentType(self.__actionId, iN), self.__inputD[iN].getContentType())
                            )
                        return False

                if self.__aReg.isSetInputFileFormat(self.__actionId, iN):
                    if self.__aReg.getInputFileFormat(self.__actionId, iN) != self.__inputD[iN].getFileFormat():
                        if self.__verbose:
                            self.__lfh.write("Failed input file format %s \n" % iN)
                        return False

        #
        numOut = self.__aReg.getOutputObjectCount(self.__actionId)
        # number of outputs consistent?
        if numOut != len(self.__outputD):
            if self.__verbose:
                self.__lfh.write("Failed number of outputs\n")
            return False
        #
        # output types consistent?
        if numOut > 0:
            for oN in self.__aReg.getOutputObjectNames(self.__actionId):
                if self.__aReg.isSetOutputReferenceType(self.__actionId, oN):
                    if self.__aReg.getOutputReferenceType(self.__actionId, oN) != self.__outputD[oN].getReferenceType():
                        if self.__verbose:
                            self.__lfh.write("Failed output reference type %s\n" % oN)
                        return False
                if self.__aReg.isSetOutputContainerType(self.__actionId, oN):
                    if self.__aReg.getOutputContainerType(self.__actionId, oN) != self.__outputD[oN].getContainerTypeName():
                        if self.__verbose:
                            self.__lfh.write("Failed output container type %s\n" % oN)
                        return False
                if self.__aReg.isSetOutputValueType(self.__actionId, oN):
                    if self.__aReg.getOutputValueType(self.__actionId, oN) != self.__outputD[oN].getValueTypeName():
                        if self.__verbose:
                            self.__lfh.write("Failed output value type %s\n" % oN)
                        return False
                if self.__aReg.isSetOutputSelectorType(self.__actionId, oN):
                    if self.__aReg.getOutputSelectorType(self.__actionId, oN) != self.__outputD[oN].getSelectorType():
                        if self.__verbose:
                            self.__lfh.write("Failed output selector type %s\n" % oN)
                        return False
                if self.__aReg.isSetOutputContentType(self.__actionId, oN):
                    if self.__aReg.getOutputContentType(self.__actionId, oN) != self.__outputD[oN].getContentType():
                        if self.__verbose:
                            self.__lfh.write("Failed output content type %s\n" % oN)
                        return False
                if self.__aReg.isSetOutputFileFormat(self.__actionId, oN):
                    if self.__aReg.getOutputFileFormat(self.__actionId, oN) != self.__outputD[oN].getFileFormat():
                        if self.__verbose:
                            self.__lfh.write("Failed output file format %s\n" % oN)
                        return False

        return True

    def run(self):
        """Invokes the specified action.

        Returns:

        The True if the action completed without exception or False otherwise.

        """
        try:
            modulePath = self.__aReg.getModuleName(self.__actionId)
            aMod = __import__(modulePath, globals(), locals(), [""])
            sys.modules[modulePath] = aMod
            #
            # Strip off any leading path to the module before we instaniate the object.
            mpL = modulePath.split(".")
            clN = mpL[-1]
            #
            # aObj = getattr(aMod,modulePath)()
            aObj = getattr(aMod, clN)(verbose=self.__verbose, log=self.__lfh)
            aMeth = getattr(aObj, self.__aReg.getMethodName(self.__actionId), None)
            if aMeth is not None:
                ok = aMeth(
                    inputObjectD=self.__inputD,
                    outputObjectD=self.__outputD,
                    userParameterD=self.__userParameterD,
                    internalParameterD=self.__aReg.getInternalParameterDict(self.__actionId),
                )
                return ok
            else:
                return False
        except Exception as _e:  # noqa: F841
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return False



##
# File:    AnnotationUtils.py
# Date:    15-Aug-2012
#
# Updates:
#  16-Aug-2012 jdw add dictionary check report
#  17-Dec-2012 jdw add option to calculation derived categories after solvent adjustment.
#  26-Jun-2013 jdw add formatCheckPdbxOp() & formatCheckPdbOp()
#  10-Oct-2013 jdw add miscCheckPdbxOp()
#  15-Jan-2014 jdw add update the content type for assemlby model files
#  16-Mar-2014 jdw add specialPositionCheckOp()
#  11-Jun-2104 jdw ad mergeXyzOp()
#  11-Jun-2104 tjo modified the check for exit status for mergeXYZop - check last line
#  14-Sep-2014 jdw add user parameter "deposit" on mergeXyzOp()
#  14-May-2015 jdw add status load method
#  20-Jan-2017 ep  add assemblyUpdateDepInfoOp()
#  15-Feb-2017 ep  add combineCifFilesOp()
##
"""
Module of annotation utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback
from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.dp.PdbxMergeCategory import PdbxMergeCategory


class ReportUtils(UtilsBase):

    """Utility class to perform annotation utility operations.

    Current supported operations include:

    - Merging of deposition ligand of interest information

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(ReportUtils, self).__init__(verbose, log)
        #  Flag to remove any temporary directories created by this class.
        #  self.__cleanUp = False
        #

    def combineLigandInfoOp(self, **kwArgs):
        """Merges the first datablock from src1 and selected categories in
        src2 and output to dst1. Will not overwrite if present
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            srcPath = inpObjD["src1"].getFilePathReference()
            mrgPath = inpObjD["src2"].getFilePathReference()
            outPath = outObjD["dst"].getFilePathReference()

            if self._verbose:
                self._lfh.write("+ReportUtils.combineLigandInfoOp() - src1 input  file paths:  %s\n" % srcPath)
                self._lfh.write("+ReportUtils.combineLigandInfoOp() - src2 input  file paths:  %s\n" % mrgPath)
                self._lfh.write("+ReportUtils.combineLigandInfoOp() - output path: %s\n" % outPath)

            pm = PdbxMergeCategory()
            # srcin, src2in, dstoit, mergelist replacelist
            ret = pm.merge(srcPath, mrgPath, outPath, ["pdbx_entry_details"], ["pdbx_binding_assay", "pdbx_entity_instance_feature"])

            if self._verbose:
                self._lfh.write("+ReportUtils.combineLigandInfoOp() - return: %s\n" % ret)

            return ret
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



##
# File:    ChemCompUtils.py
# Date:    10-Sept-2010
#
# Updates:
# 2011-09-13    RPS    Explicitly setting name of working directory used during calls for "chem-comp-assign" processing.
# 2012-06-14    jdw    Add user selection file to instance update option.
# 2013-06-27    RPS    Added chemCompAssignExactOp and chemCompAssignExactNLOp for use by LigandLite Module.
# 2014-07-07    jdw    Disable user selection file to instance update option.
##
"""
Module of chemical component utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility


class ChemCompUtils(UtilsBase):
    """Utility class to perform file format conversions.

    Current supported operations include:

    - Chemical component linkage calculation
    - Chemical component assignment
    - Chemical component instance assignment uppdate

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(ChemCompUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """
        #

    def chemCompLinkOp(self, **kwArgs):
        """Performs chemical component linkage calculation on PDBx format files."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            ccLinkPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("chem-comp-link")
            dp.exp(ccLinkPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompLinkOp() - PDBx file path:    %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompLinkOp() - CC link file path: %s\n" % ccLinkPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def chemCompAssignOp(self, **kwArgs):
        """Performs chemical component assignment calculation on PDBx format files."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()
            depDataSetId = inpObjD["src1"].getDepositionDataSetId()
            #
            ccLinkFilePath = inpObjD["src2"].getFilePathReference()
            #
            ccAssignFilePath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            ccAssignWrkngDirPath = os.path.join(dirPath, "assign")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setWorkingDir(ccAssignWrkngDirPath)
            dp.addInput(name="id", value=depDataSetId)
            if (ccLinkFilePath is not None) and os.path.exists(ccLinkFilePath):
                dp.addInput(name="cc_link_file_path", value=ccLinkFilePath, type="file")
            dp.imp(pdbxPath)
            dp.op("chem-comp-assign")
            dp.exp(ccAssignFilePath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompAssignOp() - PDBx file path:      %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompAssignOp() - CC link file path:   %s\n" % ccLinkFilePath)
                self._lfh.write("+ChemCompUtils.chemCompAssignOp() - CC assign file path: %s\n" % ccAssignFilePath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def chemCompAssignNLOp(self, **kwArgs):
        """Performs chemical component assignment calculation on PDBx format files."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            #
            ccLinkFilePath = None
            #
            ccAssignPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            ccAssignWrkngDirPath = os.path.join(dirPath, "assign")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setWorkingDir(ccAssignWrkngDirPath)
            dp.addInput(name="id", value=depDataSetId)
            if (ccLinkFilePath is not None) and os.path.exists(ccLinkFilePath):
                dp.addInput(name="cc_link_file_path", value=ccLinkFilePath, type="file")
            dp.imp(pdbxPath)
            dp.op("chem-comp-assign")
            dp.exp(ccAssignPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompAssignOp() - PDBx file path:      %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompAssignOp() - CC assign file path: %s\n" % ccAssignPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def chemCompAssignExactOp(self, **kwArgs):
        """Performs chemical component assignment calculation with exact match option on PDBx format files."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()
            depDataSetId = inpObjD["src1"].getDepositionDataSetId()
            #
            ccLinkFilePath = inpObjD["src2"].getFilePathReference()
            #
            ccAssignFilePath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            ccAssignWrkngDirPath = os.path.join(dirPath, "assign")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setWorkingDir(ccAssignWrkngDirPath)
            dp.addInput(name="id", value=depDataSetId)
            if (ccLinkFilePath is not None) and os.path.exists(ccLinkFilePath):
                dp.addInput(name="cc_link_file_path", value=ccLinkFilePath, type="file")
            dp.imp(pdbxPath)
            dp.op("chem-comp-assign-exact")
            dp.exp(ccAssignFilePath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompAssignExactOp() - PDBx file path:      %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompAssignExactOp() - CC link file path:   %s\n" % ccLinkFilePath)
                self._lfh.write("+ChemCompUtils.chemCompAssignExactOp() - CC assign file path: %s\n" % ccAssignFilePath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def chemCompAssignExactNLOp(self, **kwArgs):
        """Performs chemical component assignment calculation with exact match option on PDBx format files.
        This method assumes no link file being provided.
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            #
            ccAssignPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            ccAssignWrkngDirPath = os.path.join(dirPath, "assign")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setWorkingDir(ccAssignWrkngDirPath)
            dp.addInput(name="id", value=depDataSetId)
            #
            dp.imp(pdbxPath)
            dp.op("chem-comp-assign-exact")
            dp.exp(ccAssignPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompAssignExactNLOp() - PDBx file path:      %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompAssignExactNLOp() - CC assign file path: %s\n" % ccAssignPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def chemCompInstanceUpdateOp(self, **kwArgs):
        """Performs chemical component assignment update operations on PDBx format model files.

        Requires: extra input files for assignment and user selection.
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()
            #
            ccAssignFilePath = inpObjD["src2"].getFilePathReference()
            #
            outputModelPdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            #
            if (ccAssignFilePath is not None) and os.path.exists(ccAssignFilePath):
                dp.addInput(name="cc_assign_file_path", value=ccAssignFilePath, type="file")
            #
            # if ((ccSelectFilePath is not None)  and  os.path.exists(ccSelectFilePath)):
            #     dp.addInput(name="cc_select_file_path",value=ccSelectFilePath,type='file')
            dp.imp(pdbxPath)
            dp.op("chem-comp-instance-update")
            dp.exp(outputModelPdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+ChemCompUtils.chemCompInstanceUpdateOp() - PDBx file path:          %s\n" % pdbxPath)
                self._lfh.write("+ChemCompUtils.chemCompInstanceUpdateOp() - CC assign file path:     %s\n" % ccAssignFilePath)
                # self._lfh.write("+ChemCompUtils.chemCompInstanceUpdateOp() - CC select file path:     %s\n" % ccSelectFilePath)
                self._lfh.write("+ChemCompUtils.chemCompInstanceUpdateOp() - PDBx output file path:   %s\n" % outputModelPdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



import time
import datetime
from decimal import Decimal

# zeroDate is arbitarily 1st Jan 2000 : gives 60 year
zeroDate = datetime.datetime(2000, 1, 1, 0, 0, 0)


def getTimeZero():

    return zeroDate


def getTimeNow():

    tNow = datetime.datetime.utcnow()

    c = tNow - zeroDate
    microSec = ((c.days * 24 * 60 * 60 + c.seconds) * 1000000 + c.microseconds) / 1000000.0

    # must use Decimal(repr(time)) as python in its wisdom will only return 3dp if you are lucky
    return Decimal(repr(microSec))


def getTimeSeconds(timestamp):

    zero = time.mktime(zeroDate.timetuple())

    d = float(zero) + float(timestamp)
    return d


def getTimeString(timestamp):

    zero = time.mktime(zeroDate.timetuple())

    fmt = "%Y-%m-%d %H:%M:%S %Z"

    d = float(zero) + float(timestamp)
    return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)


def getTimeFromEpoc(timestamp):

    fmt = "%Y-%m-%d %H:%M:%S %Z"

    d = float(timestamp)
    return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)


def getTimeReadable(timestamp):

    one = ((zeroDate.days * 24 * 60 * 60 + zeroDate.seconds) * 1000000 + zeroDate.microseconds) / 1000000.0  # pylint: disable=no-member
    two = float(str(timestamp))
    return datetime.datetime.utcfromtimestamp(one + two)



"""

    File:    LocalDbApi.py

    Providing addintaional APIs for WFE to get info from local database

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "Jun 4, 2010"


"""

import sys
import MySQLdb

#
from wwpdb.utils.wf.dbapi.DbConnection import DbConnection


class LocalDbApi(object):
    """ """

    def __init__(self, log=sys.stderr, verbose=False):
        """
        connect to local database
        """
        self.__lfh = log
        self.__verbose = verbose  # pylint: disable=unused-private-member
        self.__dbServer = "mysql"
        self.__dbHost = "somehome"
        self.__dbName = "unknown"
        self.__dbUser = "user"
        self.__dbPw = "password"
        self.__dbPort = 3306

        self.__myDb = DbConnection(dbServer=self.__dbServer, dbHost=self.__dbHost, dbName=self.__dbName, dbUser=self.__dbUser, dbPw=self.__dbPw, dbPort=self.__dbPort)

        self.__dbcon = self.__myDb.connect()

    def getNewDepositedIds(self, interval):
        """
        Get newly deposited ids for RCSB site

        "interval" is a number of days from current time when the python
        code is run.

        """
        query = "select lcase(structure_id) from rcsb_status where initial_deposition_date>=DATE_SUB(curdate(), interval %s" % interval + " day) order by structure_id"
        returnList = []
        try:
            self.__dbcon.commit()
            curs = self.__dbcon.cursor()
            curs.execute(query)
            while True:
                result = curs.fetchone()
                if result is not None:
                    returnList.append(result[0])
                else:
                    break
        except MySQLdb.Error as e:
            self.__lfh.write("Database error %s: %s\n" % (e.args[0], e.args[1]))
            curs.close()
            self.__dbcon.close()
            sys.exit(1)

        return returnList



##
# File:  StatusDbApi.py
# Date:  14-Sep-2016
# Updates:
##
"""

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) 2015 wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"


import sys

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.wf.dbapi.DbApiUtil import DbApiUtil


class StatusDbApi(object):
    """ """

    __schemaMap = {  # pylint: disable=unused-private-member
        "GET_GROUP_ID": "select group_id from group_deposition_information where dep_set_id = '%s'",
        "GET_DEP_ID": "select dep_set_id from group_deposition_information where group_id = '%s'",
    }

    def __init__(self, siteId=None, verbose=False, log=sys.stderr):
        """ """
        self.__lfh = log
        self.__verbose = verbose
        self.__siteId = siteId
        self.__cI = ConfigInfo(self.__siteId)
        self.__dbServer = self.__cI.get("SITE_DB_SERVER")
        self.__dbHost = self.__cI.get("SITE_DB_HOST_NAME")
        self.__dbName = self.__cI.get("SITE_DB_DATABASE_NAME")
        self.__dbUser = self.__cI.get("SITE_DB_USER_NAME")
        self.__dbPw = self.__cI.get("SITE_DB_PASSWORD")
        self.__dbSocket = self.__cI.get("SITE_DB_SOCKET")
        self.__dbPort = int(self.__cI.get("SITE_DB_PORT_NUMBER"))
        #
        self.__dbApi = DbApiUtil(
            dbServer=self.__dbServer,
            dbHost=self.__dbHost,
            dbName=self.__dbName,
            dbUser=self.__dbUser,
            dbPw=self.__dbPw,
            dbSocket=self.__dbSocket,
            dbPort=self.__dbPort,
            verbose=self.__verbose,
            log=self.__lfh,
        )
        self.__dbApi.setSchemaMap(self.__schemaMap)

    def __getDataDir(self, key, parameter, idx):
        dlist = self.__dbApi.selectData(key=key, parameter=parameter)
        if dlist:
            return dlist[idx]
        #
        return None

    def getGroupId(self, depId=None):
        if not depId:
            return None
        #
        rtnDir = self.__getDataDir("GET_GROUP_ID", (depId), 0)
        if rtnDir:
            return rtnDir["group_id"]
        #
        return None

    def getEntryIdList(self, groupId=None):
        entryList = []
        if not groupId:
            return entryList
        #
        retList = self.__dbApi.selectData(key="GET_DEP_ID", parameter=(groupId))
        for retDir in retList:
            if "dep_set_id" in retDir and retDir["dep_set_id"]:
                entryList.append(str(retDir["dep_set_id"]))
            #
        #
        return entryList

    def runUpdate(self, table=None, where=None, data=None):
        return self.__dbApi.runUpdate(table=table, where=where, data=data)



##
# File:    AnnotationUtils.py
# Date:    15-Aug-2012
#
# Updates:
#  16-Aug-2012 jdw add dictionary check report
#  17-Dec-2012 jdw add option to calculation derived categories after solvent adjustment.
#  26-Jun-2013 jdw add formatCheckPdbxOp() & formatCheckPdbOp()
#  10-Oct-2013 jdw add miscCheckPdbxOp()
#  15-Jan-2014 jdw add update the content type for assemlby model files
#  16-Mar-2014 jdw add specialPositionCheckOp()
#  11-Jun-2104 jdw ad mergeXyzOp()
#  11-Jun-2104 tjo modified the check for exit status for mergeXYZop - check last line
#  14-Sep-2014 jdw add user parameter "deposit" on mergeXyzOp()
#  14-May-2015 jdw add status load method
#  20-Jan-2017 ep  add assemblyUpdateDepInfoOp()
#  15-Feb-2017 ep  add combineCifFilesOp()
##
"""
Module of annotation utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import logging
import os
import shutil
import socket
import sys
import tempfile
import traceback

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.dp.DensityWrapper import DensityWrapper
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase

try:
    # We will have present on annotation system - but allow testing without
    from wwpdb.apps.ann_tasks_v2.io.PisaReader import PisaAssemblyReader
    from wwpdb.apps.ann_tasks_v2.em3d.EmAutoFix import EmAutoFix
    from wwpdb.apps.ann_tasks_v2.em3d.EmMapAutoFixVers import EmMapAutoFixVers
    from wwpdb.apps.ann_tasks_v2.em3d.EmHeaderUtils import EmHeaderUtils
    from wwpdb.apps.ann_tasks_v2.related.UpdateRelated import UpdateRelated
    from wwpdb.apps.ann_tasks_v2.expIoUtils.PdbxExpUpdate import PdbxExpUpdate
    from wwpdb.utils.session.WebRequest import InputRequest
    from wwpdb.io.locator.PathInfo import PathInfo
except ImportError:
    pass

from wwpdb.utils.db.DbLoadingApi import DbLoadingApi
from mmcif.io.IoAdapterCore import IoAdapterCore

logger = logging.getLogger(__name__)


class AnnotationUtils(UtilsBase):
    """Utility class to perform annotation utility operations.

    Current supported operations include:

    - Assembly calculations

    - Secondary structure
    - Site environment
    - Solvent reorganization
    - Nucleic acid geometrical features
    - Chemical linkages and cis-peptide linkages
    - Automatic running of mapfix on archive files
    - Automatic lookup of pdbx_database_related

    -Various format, dictionary and geometry checks

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(AnnotationUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """
        #

    def statusLoadOp(self, **kwArgs):
        """Performs PDBx database load of status related categories."""
        try:
            (inpObjD, _outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            dirPath = inpObjD["src"].getDirPathReference()
            #
            if self._verbose:
                self._lfh.write("+AnnotationUtils.statusLoadOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.statusLoadOp() - deposition ID:          %s\n" % depDataSetId)
            dbLd = DbLoadingApi(log=self._lfh, verbose=self._verbose)
            ok = dbLd.doLoadStatus(pdbxPath, dirPath)
            self._lfh.write("+AnnotationUtils.statusLoadOp() - returns   %r\n" % ok)
            return ok
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def specialPositionCheckOp(self, **kwArgs):
        """Performs special position check and produces a summary report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "special-position-check.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("annot-dcc-special-position")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.specialPositionCheckOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.specialPositionCheckOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def formatCheckPdbxOp(self, **kwArgs):
        """Performs PDBx format check and returns a text check report."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "format-check-pdbx.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            if "nmr_flag" in uD:
                isNmr = str(uD["nmr_flag"])
                if isNmr == "Y":
                    dp.addInput(name="nmr", value="Y")
            #
            dp.op("annot-format-check-pdbx")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.formatCheckPdbxOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.formatCheckPdbxOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def formatCheckPdbOp(self, **kwArgs):
        """Performs PDB format check and returns a text check report."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            pdbPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "format-check-pdb.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbPath)
            if "nmr_flag" in uD:
                isNmr = str(uD["nmr_flag"])
                if isNmr == "Y":
                    dp.addInput(name="nmr", value="Y")

            #
            dp.op("annot-format-check-pdb")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.formatCheckPdbOp() - PDB input  file path:  %s\n" % pdbPath)
                self._lfh.write("+AnnotationUtils.formatCheckPdbOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def dictCheckOp(self, **kwArgs):
        """Performs PDBx dictionary check on PDBx format input file and returns a text check report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "dict-check.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("check-cif")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.dictCheckOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.dictCheckOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def dictCheckFirstOp(self, **kwArgs):
        """Performs PDBx dictionary check on the first block of a PDBx format input file
        and returns a text check report.
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "dict-check.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)

            dp.addInput("first_block")

            #
            dp.op("check-cif")
            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.dictCheckOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.dictCheckOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def dictR4CheckOp(self, **kwArgs):
        """Performs PDBx R4 dictionary check on PDBx format input file and returns a text check report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "dict-check.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("check-cif-v4")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.dictR4CheckOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.dictR4CheckOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def geometryCheckOp(self, **kwArgs):
        """Performs geometry check on PDBx format input file and returns a text check report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "geometry-check-report.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("validate-geometry")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.geometryCheckOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.geometryCheckOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def miscCheckReportOp(self, **kwArgs):
        """Performs additional checks on PDBx format model files and returns a text check report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "misc-check-report.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("annot-extra-checks")

            dp.expLog(logPath)
            dp.exp(reportPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.miscCheckReportOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.miscCheckReportOp() - Report file path:       %s\n" % reportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def dccCheckReportOp(self, **kwArgs):
        """DCC generated check report of model and SF file --"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            sfPath = inpObjD["src2"].getFilePathReference()
            xyzPath = inpObjD["src1"].getFilePathReference()
            #
            #
            dirPath = outObjD["dst1"].getDirPathReference()
            reportFilePath = outObjD["dst1"].getFilePathReference()
            logFilePath = outObjD["dst2"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(xyzPath)
            if (sfPath is not None) and os.path.exists(sfPath):
                dp.addInput(name="sf_file_path", value=sfPath, type="file")

            dp.op("annot-dcc-report")
            dp.expLog(logFilePath)
            dp.exp(reportFilePath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.dccCheckReportOp() - SF  input  file path:          %s\n" % sfPath)
                self._lfh.write("+FormatUtils.dccCheckReportOp() - XYZ input  file path:          %s\n" % xyzPath)
                self._lfh.write("+FormatUtils.dccCheckReportOp() - LOG output file path:          %s\n" % logFilePath)
                self._lfh.write("+FormatUtils.dccCheckReportOp() - REPORT/PDBx output file path:  %s\n" % reportFilePath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def nucleicAcidGeometryOp(self, **kwArgs):
        """Performs base pair/step geometry calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "base-pair-position.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("annot-base-pair-info")

            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.nucleicAcidGeometryOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.nucleicAcidGeometryOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def solventPositionAddDerivedOp(self, **kwArgs):
        """Performs solvent reorganization calculation on PDBx format input file and update this data in the PDBx model file.

        This version also adds selected derived categories required by other annotation tasks.   This step should
        precede other added annotation tasks.

        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "solvent-position.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            # if (solventArgs is not None):
            #    dp.addInput(name="solvent_arguments",value=solventArgs)
            #
            dp.op("annot-reposition-solvent-add-derived")
            # dp.op("annot-distant-solvent")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.solventPositionOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.solventPositionOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def solventPositionOp(self, **kwArgs):
        """Performs solvent reorganization calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "solvent-position.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            # if (solventArgs is not None):
            #    dp.addInput(name="solvent_arguments",value=solventArgs)
            #
            dp.op("annot-reposition-solvent")
            # dp.op("annot-distant-solvent")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.solventPositionOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.solventPositionOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def solventPositionAnalysis(self, **kwArgs):
        """Performs distant solvent calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "solvent-position-analysis.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            # if (solventArgs is not None):
            #    dp.addInput(name="solvent_arguments",value=solventArgs)
            #
            dp.op("annot-distant-solvent")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.solventPositionAnalysis() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.solventPositionAnalysis() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def linkOp(self, **kwArgs):
        """Performs link calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "annot-link-and-ss-bond.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-link-ssbond")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.linkOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.linkOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def cisPeptideOp(self, **kwArgs):
        """Performs cis-peptide link perception on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-cis-peptide")
            # dp.expLog("annot-cis-peptide.log")
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.cisPeptideOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.cidPeptideOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updateGeometryValidationOp(self, **kwArgs):
        """Updates geometrical validation diagnostics in the PDBx model file -"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-validate-geometry")
            # dp.expLog("annot-validate-geometry.log")
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.updateGeometryValidationOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.updateGeometryValidationOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def getCorresInfoOp(self, **kwArgs):
        """gets correspondance information for the PDBx model file -"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-get-corres-info")
            # dp.expLog("annot-get-corres-info")
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.getCorresInfo() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.getCorresInfo() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def secondaryStructureOp(self, **kwArgs):
        """Performs secondary structure calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "annot-secondary-structure-w-top.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            # topPath=
            # dp.addInput(name="ss_topology_file_path",value=topPath)
            dp.op("annot-secondary-structure")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def secondaryStructureTopologyOp(self, **kwArgs):
        """Performs secondary structure calculation on PDBx format input file and update this data in the PDBx model file.

        This method includes the option for a supplementary topology file -

        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()

            topFilePath = inpObjD["src2"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "annot-secondary-structure-w-top.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.addInput(name="ss_topology_file_path", value=topFilePath)
            dp.op("annot-secondary-structure")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx input  file path:     %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - Topology input file path:  %s\n" % topFilePath)
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx output file path:     %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def consolidatedTasksOp(self, **kwArgs):
        """Performs most annotation tasks as a single step -

        This method includes the option for a supplementary topology file -

        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()

            topFilePath = inpObjD["src2"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "annot-consolidated-tasks-w-top.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.addInput(name="ss_topology_file_path", value=topFilePath)
            dp.op("annot-consolidated-tasks")
            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx input  file path:     %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - Topology input file path:  %s\n" % topFilePath)
                self._lfh.write("+AnnotationUtils.secondaryStructureOp() - PDBx output file path:     %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def siteEnvironmentOp(self, **kwArgs):
        """Performs site environment calculation on PDBx format input file and update this data in the PDBx model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "site-anal.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            # Step 1
            resultPath = os.path.join(dirPath, depDataSetId + "-site-anal.cif")
            dp.imp(pdbxPath)
            dp.addInput(name="block_id", value=depDataSetId)
            dp.op("annot-site")
            dp.expLog(logPath)
            dp.exp(resultPath)

            # Step 2
            dp.imp(pdbxPath)
            dp.addInput(name="site_info_file_path", value=resultPath, type="file")
            dp.op("annot-merge-struct-site")
            #
            # dp.expLog(logPath2)
            dp.exp(pdbxOutputPath)
            #
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.siteEnvironmentOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.siteEnvironmentOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
                self._lfh.write("+AnnotationUtils.siteEnvironmentOp() - Site anal file path:    %s\n" % resultPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def assemblyReportOp(self, **kwArgs):
        """Performs assembly prediction on PDBx format input files and produces an assembly report."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()

            assemblyReportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            pisaSession = "session_" + depDataSetId
            logPath1 = os.path.join(dirPath, "assembly-analysis.log")
            logPath2 = os.path.join(dirPath, "assembly-report.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.addInput(name="pisa_session_name", value=pisaSession)
            dp.op("pisa-analysis")
            dp.expLog(logPath1)
            dp.op("pisa-assembly-report-xml")
            dp.exp(assemblyReportPath)
            dp.expLog(logPath2)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.chemCompLinkOp() - PDBx file path:    %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.chemCompLinkOp() - Assembly report file path: %s\n" % assemblyReportPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def assemblyReportAndModelsOp(self, **kwArgs):
        """Performs assembly prediction on PDBx format input files and produces an assembly report.

        The list of file paths for the materialized model coordinates for each assembly is also returned.

        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            assemblyReportPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            pisaSession = "session_" + depDataSetId
            logPath1 = os.path.join(dirPath, "assembly-analysis.log")
            logPath2 = os.path.join(dirPath, "assembly-report.log")
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            if self._verbose:
                self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - PDBx file path:    %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - Assembly report file path: %s\n" % assemblyReportPath)
                self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - dirPath : %s\n" % dirPath)
                self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - siteId  : %s\n" % siteId)
                self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - logPath : %s\n" % logPath1)
            #
            maxAssems = 50

            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.addInput(name="pisa_session_name", value=pisaSession)
            dp.op("pisa-analysis")
            dp.expLog(logPath1)
            dp.op("pisa-assembly-report-xml")
            dp.exp(assemblyReportPath)
            dp.expLog(logPath2)

            #
            assemD = self.__readAssemblyReport(assemblyReportPath)
            pathList = []
            if len(assemD) > 0:
                if self._verbose:
                    self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - assembly uid list %r\n" % assemD.keys())
                #
                for assemblyUid in sorted(assemD.keys()):
                    if assemblyUid > maxAssems:
                        break
                    if assemblyUid == 0:
                        continue
                    assemModelFileName = os.path.join(dirPath, depDataSetId + "_assembly-model-xyz_P" + str(assemblyUid) + ".cif.V1")
                    dp.addInput(name="pisa_assembly_id", value=assemblyUid)
                    dp.op("pisa-assembly-coordinates-cif")
                    dp.exp(assemModelFileName)
                    if self._verbose:
                        self._lfh.write("+AnnotationUtils.assemblyReportAndModels() - creating assembly model %r file %s\n" % (assemblyUid, assemModelFileName))
                    pathList.append(assemModelFileName)

            outObjD["dst2"].setValue(pathList)
            if self.__cleanUp:
                dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    # def assemblyInstanceUpdateOp(self, **kwArgs):
    #     """Performs assembly assignment update operations on PDBx format model files.

    #        Requires: extra input files for assignment and user selection.
    #     """
    #     try:
    #         (inpObjD, outObjD, uD, pD) = self._getArgs(kwArgs)
    #         pdbxPath = inpObjD["src1"].getFilePathReference()
    #         #
    #         reportFilePath = inpObjD["src2"].getFilePathReference()
    #         selectFilePath = inpObjD["src3"].getFilePathReference()
    #         #
    #         outputModelPdbxPath = outObjD["dst"].getFilePathReference()
    #         dirPath = outObjD["dst"].getDirPathReference()
    #         logPath = os.path.join(dirPath, "assembly-update.log")
    #         #
    #         cI = ConfigInfo()
    #         siteId = cI.get("SITE_PREFIX")
    #         dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
    #         #
    #         dp.imp(pdbxPath)
    #         selectString = __readSelection(selectFilePath)
    #         iL = []
    #         sL = selectString.split(',')
    #         for s in sL:
    #             if s is not None and len(s) > 0:
    #                 iL.append("%d" % int(s))

    #         tupleList = ','.join(iL)
    #         dp.addInput(name="pisa_assembly_tuple_list", value=tupleList)
    #         dp.addInput(name="pisa_assembly_file_path", value=reportFilePath)
    #         #
    #         dp.op("pisa-assembly-merge-cif")
    #         dp.exp(outputModelPdbxPath)
    #         dp.expLog(logPath)

    #         if (self._verbose):
    #             self._lfh.write("+AnnotationUtils.assemblyInstanceUpdateOp() - PDBx file path:         %s\n" % pdbxPath)
    #             self._lfh.write("+AnnotationUtils.assemblyInstanceUpdateOp() - report file path:       %s\n" % reportFilePath)
    #             self._lfh.write("+AnnotationUtils.assemblyInstanceUpdateOp() - select file path:       %s\n" % selectFilePath)
    #             self._lfh.write("+AnnotationUtils.assemblyInstanceUpdateOp() - PDBx output file path:  %s\n" % outputModelPdbxPath)

    #         if (self.__cleanUp):
    #             dp.cleanup()
    #         return True
    #     except Exception as _e:  # noqa: F841
    #         traceback.print_exc(file=self._lfh)
    #         return False

    def assemblyUpdateDepInfoOp(self, **kwArgs):
        """Performs an update of the pdbx_struct_assembly_gen_depositor_info"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()

            pdbxOutputPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "assembly-depinfo-update.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            #
            dp.op("annot-update-dep-assembly-info")

            dp.expLog(logPath)
            dp.exp(pdbxOutputPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.assemblyUpdateDepInfoOp() - PDBx input  file path:  %s\n" % pdbxPath)
                self._lfh.write("+AnnotationUtils.assemblyUpdateDepInfoOp() - PDBx output file path:  %s\n" % pdbxOutputPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def __readAssemblyReport(self, reportPath):
        """Read assembly calculation report and return a dictionary of assembly details."""
        pA = PisaAssemblyReader(verbose=self._verbose, log=self._lfh)
        _ok = pA.read(reportPath)  # noqa: F841
        self.__assemblyD = pA.getAssemblyDict()  # pylint: disable=attribute-defined-outside-init
        return self.__assemblyD

    # def __readSelection(self, selectPath):
    #     """Read the comma separated selection string from a file in the current session directory."""
    #     try:
    #         ofh = open(selectPath, "r")
    #         tS = ofh.readline()
    #         ofh.close()
    #         if (tS is not None) and (len(tS) > 1):
    #             return tS[:-1]
    #         else:
    #             return None
    #     except Exception as _e:  # noqa: F841
    #         if self._verbose:
    #             traceback.print_exc(file=self._lfh)
    #         return None

    def mergeXyzOp(self, **kwArgs):
        """Merge coordinate data files and report status -"""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inpXyzPath = inpObjD["src1"].getFilePathReference()
            newXyzPath = inpObjD["src2"].getFilePathReference()

            if "new_coordinate_format" in uD:
                iFormat = str(uD["new_coordinate_format"])
            else:
                iFormat = "cif"

            if iFormat in ["cif", "pdbx", "mmcif"]:
                newXyzFormat = "cif"
            else:
                newXyzFormat = "pdb"

            isDeposit = False
            if "deposit" in uD and (str(uD["deposit"]).lower() != "no"):
                isDeposit = True

            #
            dirPath = outObjD["dst1"].getDirPathReference()
            mergedFilePath = outObjD["dst1"].getFilePathReference()
            logFilePath = outObjD["dst2"].getFilePathReference()

            if self._verbose:
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ input   file path:  %s\n" % inpXyzPath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ new     file path:  %s\n" % newXyzPath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ new file   format:  %s\n" % newXyzFormat)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - directory path:         %s\n" % dirPath)
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setDebugMode(flag=True)
            dp.imp(inpXyzPath)
            if (newXyzPath is not None) and os.path.exists(newXyzPath):
                dp.addInput(name="new_coordinate_file_path", value=newXyzPath, type="file")
                dp.addInput(name="new_coordinate_format", value=newXyzFormat, type="param")
            else:
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ new source input not found in file path:  %s\n" % newXyzPath)
            if isDeposit:
                dp.addInput(name="deposit", value=True, type="param")

            dp.op("annot-merge-xyz")
            dp.expLog(logFilePath)
            dp.exp(mergedFilePath)

            if not os.access(mergedFilePath, os.R_OK):
                myStatus = "error"
            else:
                myStatus = self.__checkMergeStatus(logFilePath)

            outObjD["dst3"].setValue(myStatus)

            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ input   file path:  %s\n" % inpXyzPath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - XYZ new     file path:  %s\n" % newXyzPath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - LOG output  file path:  %s\n" % logFilePath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - PDBx merged file path:  %s\n" % mergedFilePath)
                self._lfh.write("+AnnotationUtils.mergeXyzOp() - return status:          %r\n" % myStatus)

            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def __checkMergeStatus(self, logFilePath):
        # Tom changed this to check the last line as finished!
        # if the first line is not 'finished!' then there is a failure -
        status = "error"
        if os.access(logFilePath, os.R_OK):
            ifh = open(logFilePath, "r")
            for line in ifh:
                if str(line).upper().startswith("FINISHED"):
                    status = "ok"

            ifh.close()
        else:
            status = "error"
        return status

    def combineCifFilesOp(self, **kwArgs):
        """Merges the first datablock from src1 and selected categories in
        src2 and output to dst1. Will not overwrite if present
        """
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            srcPath = inpObjD["src1"].getFilePathReference()
            mrgPath = inpObjD["src2"].getFilePathReference()
            outPath = outObjD["dst"].getFilePathReference()
            if "categories" in uD:
                # Handle incomming unicode
                mrgCat = str(uD["categories"]).split(",")
            else:
                mrgCat = None

            if self._verbose:
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - src1 input  file paths:  %s\n" % srcPath)
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - src2 input  file paths:  %s\n" % mrgPath)
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - output path: %s\n" % outPath)
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - mrgCat: %s\n" % mrgCat)

            ioObj = IoAdapterCore(verbose=self._verbose, log=self._lfh)
            tmpdir = os.path.abspath(os.path.dirname(srcPath))
            dIn = ioObj.readFile(inputFilePath=srcPath, outDirPath=tmpdir)
            if not dIn:
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() -failed to load src1\n")
                return False

            srcBlock = dIn[0]

            mrgIn = ioObj.readFile(inputFilePath=mrgPath, selectList=mrgCat, outDirPath=tmpdir)
            if not mrgIn:
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - Failed to load src2\n")
                return False

            mrgBlock = mrgIn[0]

            dstNameList = srcBlock.getObjNameList()

            if len(mrgBlock.getObjNameList()) == 0:
                self._lfh.write("+AnnotationUtils.combineCifFilesOp() - nothing to merge\n")
                return True

            for cName in mrgBlock.getObjNameList():
                # If destination has object - do not overwrite
                if cName not in dstNameList:
                    cObj = mrgBlock.getObj(cName)
                    srcBlock.append(cObj)

            # Write out
            ret = ioObj.writeFile(outputFilePath=outPath, containerList=dIn)
            return ret
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def em3dAutoMapFixOp(self, **kwArgs):
        """Performs mapfix if needed automatically.  Typically run on model and map files in archive as they need to be in sync."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            sessdir = inpObjD["sessdir"].getDirPathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            pdbxOutPath = outObjD["dst"].getFilePathReference()

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            eaf = EmAutoFix(sessionPath=sessdir, siteId=siteId)
            _ret = eaf.autoFixMapLabels(datasetid=depDataSetId, modelin=pdbxPath, modelout=pdbxOutPath)  # noqa: F841
            # Always return true - even if no work done
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def em3dAutoEmMapFixVersOp(self, **kwArgs):
        """Updates version numbers of files in em_map category in model file.  Typically run on model and map files in archive as they need to be in sync."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            sessdir = inpObjD["sessdir"].getDirPathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            pdbxOutPath = outObjD["dst"].getFilePathReference()

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            if "location" in uD:
                location = str(uD["location"])
            else:
                location = "archive"

            eaf = EmMapAutoFixVers(sessionPath=sessdir, siteId=siteId)
            ret = eaf.autoFixEmMapVersions(datasetid=depDataSetId, modelin=pdbxPath, modelout=pdbxOutPath, location=location)
            self._lfh.write("+em3dAutoEmMapFixVersOp fixvers returns %s\n" % ret)

            # Always return true - even if no work done
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updatePdbxDatabaseRelatedOp(self, **kwArgs):
        """Updates pdbx_database_related by performing lookups on deposition ids D_XXX in the db_id field for PDB, BMRB and EMDB"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            pdbxOutPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            logPath = os.path.join(dirPath, "update_pdbx_database_related.log")

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            ur = UpdateRelated(siteId=siteId)
            ret = ur.updateRelatedEntries(pdbxPath, pdbxOutPath, logPath)

            self._lfh.write("+updatePdbxDatabaseRelatedOp returns %s\n" % ret)

            # Always return true - even if no work done
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updateSFWavelengthOp(self, **kwArgs):
        """Updates wavelnegth in SF fie"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()
            sfPath = inpObjD["src2"].getFilePathReference()
            sfOutPath = outObjD["dst"].getFilePathReference()
            # For session directory
            dirPath = inpObjD["src1"].getDirPathReference()
            entryId = inpObjD["src1"].getDepositionDataSetId()

            if not os.path.exists(sfPath):
                # NMR, EM, etc
                return True

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            myReqObj = InputRequest({}, verbose=True, log=sys.stderr)
            myReqObj.setValue("TopSessionPath", dirPath)
            myReqObj.setValue("WWPDB_SITE_ID", siteId)
            myReqObj.newSessionObj()

            peu = PdbxExpUpdate(myReqObj, verbose=self._verbose, log=self._lfh)
            ret = peu.doUpdate(entryId, modelInputFile=pdbxPath, expInputFile=sfPath, expOutputFile=sfOutPath, skipNotChanged=True)

            self._lfh.write("+updateSFWavelengthOp returns %s\n" % ret)

            # Always return true - even if no work done
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def emdXmlHeaderCheckOp(self, **kwArgs):
        """Checks EMD -> XML header conversion"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            depDataSetId = inpObjD["src"].getDepositionDataSetId()

            reportPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            # logPath = os.path.join(dirPath, "dict-check.log")

            ioObj = IoAdapterCore(verbose=self._verbose, log=self._lfh)
            dIn = ioObj.readFile(inputFilePath=pdbxPath, selectList=["em_admin"])
            if not dIn or len(dIn) == 0:
                return True

            cObj = dIn[0].getObj("em_admin")
            if not cObj:
                # No em_admin
                return True

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            hostName = str(socket.gethostname()).split(".")[0]  # pylint: disable=no-member
            if (hostName is not None) and (len(hostName) > 0):
                suffix = "-" + hostName
            else:
                suffix = "-dir"
            wrkPath = tempfile.mkdtemp(suffix, "rcsb-", dirPath)
            os.chmod(wrkPath, 0o750)

            emdXmlPath = os.path.join(wrkPath, depDataSetId + "-emd.xml")

            fPath = pdbxPath

            emh = EmHeaderUtils(siteId=siteId, verbose=self._verbose, log=self._lfh)
            if hasattr(emh, "transEmd"):
                # Still running EM->EMD conversion
                emdModelPath = os.path.join(wrkPath, depDataSetId + "_model-emd.cif")
                status = emh.transEmd(pdbxPath, emdModelPath, mode="src-dst", tags=[])
                if not status:
                    self._lfh.write("Translation of model failed\n")
                    return True
                fPath = emdModelPath

            status = emh.transHeader(fPath, emdXmlPath, reportPath, validateXml=True)
            self._lfh.write("Status of xml translation %s\n" % status)

            # Cleanup directory
            shutil.rmtree(wrkPath, ignore_errors=True)

            # Always return true - even if no work done
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def emVolumeBcifConversionOp(self, **kwArgs):
        """
        converts EM volume from map format into bcif
        :param kwArgs:
        :return bool: True if worked, False if failed
        """
        try:
            (inpObjD, _outObjD, _uD, _pD) = self._getArgs(kwArgs)
            modelPath = inpObjD["src"].getFilePathReference()
            # depDataSetId = inpObjD["src"].getDepositionDataSetId()

            dirPath = inpObjD["src"].getDirPathReference()
            if not os.path.exists(modelPath):
                # no model
                raise IOError("Missing model file")

            ioObj = IoAdapterCore(verbose=self._verbose, log=self._lfh)
            dIn = ioObj.readFile(inputFilePath=modelPath, selectList=["em_map"])
            if not dIn or len(dIn) == 0:
                return True

            cObj = dIn[0].getObj("em_map")
            if not cObj:
                # No em_map
                return True

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            pi = PathInfo(siteId=siteId)

            # loop through all the map file names in the mmcif file and convert to Bcif files
            for mapNumber in range(0, len(cObj)):
                mapName = cObj.getValue("file", mapNumber)
                mapNameInfo = pi.parseFileName(mapName)
                mapPath = pi.getFilePath(mapNameInfo[0], contentType=mapNameInfo[1], formatType=mapNameInfo[2], partNumber=mapNameInfo[3])
                mapBcifPath = pi.getFilePath(mapNameInfo[0], contentType=mapNameInfo[1], formatType="bcif", partNumber=mapNameInfo[3])
                dw = DensityWrapper()
                dw.convert_em_volume(in_em_volume=mapPath, out_binary_volume=mapBcifPath, working_dir=dirPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def xrayBcifConversionOp(self, **kwArgs):
        """
        converts X-ray maps from maps format into bcif
        :param kwArgs:
        :return bool: True if worked, False if failed
        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            coordinates = inpObjD["src1"].getFilePathReference()
            twofofcmap = inpObjD["src2"].getFilePathReference()
            fofcmap = inpObjD["src3"].getFilePathReference()
            # depDataSetId = inpObjD["src1"].getDepositionDataSetId()

            mapBcifPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            if os.path.exists(twofofcmap) and os.path.exists(fofcmap) and os.path.exists(coordinates):
                dw = DensityWrapper()
                return dw.convert_xray_density_map(coord_file=coordinates, in_2fofc_cif=twofofcmap, in_fofc_cif=fofcmap, out_binary_volume=mapBcifPath, working_dir=dirPath)
            else:
                # no x-ray mmCIF map files
                return True

        except Exception as _e:  # noqa: F841
            logging.error(_e)
            traceback.print_exc(file=self._lfh)
            return False



##
# File:    SeqStatsUtils.py
# Date:    29-April-2010
#
# Updates:
#
#   22-March-2013 jdw obsolete -
##
"""
Module of sequence processing utilities.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase


class SeqStatsUtils(UtilsBase):
    """Utility class to perform sequence processing operations.

    Current supported operations include:


    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=True, log=sys.stderr):
        super(SeqStatsUtils, self).__init__(verbose, log)
        # Flag to remove any temporary directories created by this class.
        self.__cleanUp = True  # pylint: disable=unused-private-member
        #

    def prepareSequenceDataOp(self, **kwArgs):
        """Prepare sequence data for alignment and selection operations.

        Model file and seqdb-match data files must all be in the source directory.
        Latest versions of these files will be used.
        """
        try:
            (_inpObjD, _outObjD, _uD, _pD) = self._getArgs(kwArgs)

            # ------
            #  This provide all information to define source data --
            #
            if self._verbose:
                self._lfh.write("+SeqStatsUtils.prepareSequenceDataOp()  This module is obsolete --- \n")
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def evalSequenceDataOp(self, **kwArgs):
        """Assess sequence data and alignment against correspondence criteria.

        Returns:

        True for corresponding sequence data or False otherwise.

        """
        try:
            (_inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)

            # ------
            #  This provide all information to define source data --
            testValue = False
            #
            if self._verbose:
                self._lfh.write("+SeqStatsUtils() assesSequenceDataOp return : %r\n" % testValue)
            outObjD["dst"].setValue(testValue)

            return True
            #

        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updateModelSequenceOp(self, **kwArgs):
        """Update the model sequence with alignment edits and selection data.

        Returns:

        True for success or False otherwise.

        """
        try:
            (_inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)

            # ------
            #  This provide all information to define source data --
            #
            #
            #
            wfInstanceId = outObjD["dst1"].getWorkflowInstanceId()
            depDataSetId = outObjD["dst1"].getDepositionDataSetId()
            fileSource = outObjD["dst1"].getStorageType()
            outModelPath = outObjD["dst1"].getDirPathReference()
            if self._verbose:
                self._lfh.write("+SeqStatsUtils() updateModelSequenceOp instance    : %s\n" % outModelPath)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceOp instance id : %s\n" % wfInstanceId)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceOp dep data id : %s\n" % depDataSetId)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceOp fileSoure   : %s\n" % fileSource)

            if self._verbose:
                self._lfh.write("+SeqStatsUtils.updateModelSequenceOp()  This module is obsolete --- \n")
            return True
            #

        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updateModelSequenceAssignOp(self, **kwArgs):
        """Update model coordinate file with aligmnent mapping information.

        Returns:

        True for success or False otherwise.

        """
        try:
            (_inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)

            # ------
            #  This provide all information to define source data --
            #
            wfInstanceId = outObjD["dst1"].getWorkflowInstanceId()
            depDataSetId = outObjD["dst1"].getDepositionDataSetId()
            fileSource = outObjD["dst1"].getStorageType()
            outModelPath = outObjD["dst1"].getDirPathReference()
            if self._verbose:
                self._lfh.write("+SeqStatsUtils() updateModelSequenceAssignOp instance    : %s\n" % outModelPath)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceAssignOp instance id : %s\n" % wfInstanceId)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceAssignOp dep data id : %s\n" % depDataSetId)
                self._lfh.write("+SeqStatsUtils() updateModelSequenceAssignOp fileSoure   : %s\n" % fileSource)

            if self._verbose:
                self._lfh.write("+SeqStatsUtils.updateModelSequenceAssignOp() This module is obsolete --- \n")

            return True
            #

        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



"""
   File:    WfSchemaMap.py
   A data class containing schema definitions for WF Database.

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"

"""


class WfSchemaMap(object):
    _schemaMap = {
        "DEPOSITION": {
            "ATTRIBUTES": {
                "DEP_SET_ID": "dep_set_id",
                "PDB_ID": "pdb_id",
                "INITIAL_DEPOSITION_DATE": "initial_deposition_date",
                "ANNOTATOR_INITIALS": "annotator_initials",
                "DEPOSIT_SITE": "deposit_site",
                "PROCESS_SITE": "process_site",
                "STATUS_CODE": "status_code",
                "AUTHOR_RELEASE_STATUS_CODE": "author_release_status_code",
                "TITLE": "title",
                "AUTHOR_LIST": "author_list",
                "EXP_METHOD": "exp_method",
                "STATUS_CODE_EXP": "status_code_exp",
                "SG_CENTER": "sg_center",
                "TITLE_EMDB": "title_emdb",
                "AUTHOR_LIST_EMDB": "author_list_emdb",
                "DEP_AUTHOR_RELEASE_STATUS_CODE_EMDB": "dep_author_release_status_code_emdb",
                "STATUS_CODE_EMDB": "status_code_emdb",
                "DATE_BEGIN_PROCESSING": "date_begin_processing",
                "DATE_END_PROCESSING": "date_end_processing",
                "DEPPW": "depPW",
                "NOTIFY": "notify",
                "EMAIL": "email",
                "LOCKING": "locking",
                "COUNTRY": "country",
                "NMOLECULE": "nmolecule",
                "EMDB_ID": "emdb_id",
                "BMRB_ID": "bmrb_id",
                "STATUS_CODE_BMRB": "status_code_bmrb",
                "STATUS_CODE_OTHER": "status_code_other",
                "POST_REL_STATUS": "post_rel_status",
                "POST_REL_RECVD_COORD": "post_rel_recvd_coord",
                "POST_REL_RECVD_COORD_DATE": "post_rel_recvd_coord_date",
            },
            "TABLE_NAME": "deposition",
        },
        "DATABASE_REF": {"ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "DATABASE_NAME": "database_name", "DATABASE_CODE": "database_code"}, "TABLE_NAME": "database_ref"},
        "WF_TASK": {
            "ATTRIBUTES": {
                "ORDINAL": "ordinal",
                "WF_TASK_ID": "wf_task_id",
                "WF_INST_ID": "wf_inst_id",
                "WF_CLASS_ID": "wf_class_id",
                "DEP_SET_ID": "dep_set_id",
                "TASK_NAME": "task_name",
                "TASK_STATUS": "task_status",
                "STATUS_TIMESTAMP": "status_timestamp",
                "TASK_TYPE": "task_type",
            },
            "TABLE_NAME": "wf_task",
        },
        "WF_INSTANCE": {
            "ATTRIBUTES": {
                "ORDINAL": "ordinal",
                "WF_INST_ID": "wf_inst_id",
                "WF_CLASS_ID": "wf_class_id",
                "DEP_SET_ID": "dep_set_id",
                "OWNER": "owner",
                "INST_STATUS": "inst_status",
                "STATUS_TIMESTAMP": "status_timestamp",
            },
            "TABLE_NAME": "wf_instance",
        },
        "WF_REFERENCE": {
            "ATTRIBUTES": {
                "DEP_SET_ID": "dep_set_id",
                "WF_INST_ID": "wf_inst_id",
                "WF_TASK_ID": "wf_task_id",
                "WF_CLASS_ID": "wf_class_id",
                "HASH_ID": "hash_id",
                "VALUE": "value",
            },
            "TABLE_NAME": "wf_reference",
        },
        "WF_CLASS_DICT": {
            "ATTRIBUTES": {
                "WF_CLASS_ID": "wf_class_id",
                "WF_CLASS_NAME": "wf_class_name",
                "WF_CLASS_FILE": "class_file",
                "TITLE": "title",
                "AUTHOR": "author",
                "VERSION": "version",
            },
            "TABLE_NAME": "wf_class_dict",
        },
        "PROCESS_INFORMATION": {
            "ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "SERIAL_NUMBER": "serial_number", "PROCESS_BEGIN": "process_begin", "PROCESS_END": "process_end", "REMARK": "remark"},
            "TABLE_NAME": "process_information",
        },
        "DA_USERS": {
            "ATTRIBUTES": {"USER_NAME": "user_name", "PASSWORD": "password", "GROUPNAME": "groupname", "EMAIL": "email", "INITIALS": "initials"},
            "TABLE_NAME": "da_users",
        },
        "DATABASE_RELATED": {
            "ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "DB_NAME": "db_name", "DETAILS": "details", "CONTENT_TYPE": "content_type", "DB_ID": "db_id"},
            "TABLE_NAME": "database_related",
        },
        "DATABASE_PDB_OBS_SPR": {
            "ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "ID": "id", "DATE": "date", "PDB_ID": "pdb_id", "REPLACE_PDB_ID": "replace_pdb_id"},
            "TABLE_NAME": "database_PDB_obs_spr",
        },
        "AUTHOR_CORRECTIONS": {
            "ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "CORRECTIONS": "corrections", "SENDING_DATE": "sending_date", "REMARK": "content_type"},
            "TABLE_NAME": "author_corrections",
        },
        "DEP_WITH_PROBLEMS": {"ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "PROBLEM_TYPE": "problem_type", "PROBLEM_DETAILS": "problem_details"}, "TABLE_NAME": "dep_with_problems"},
        "RELEASE_REQUEST": {
            "ATTRIBUTES": {"DEP_SET_ID": "dep_set_id", "REQ_CITATION": "req_citation", "RELEASE_DATE": "release_date", "PUBMED_ID": "PubMed_id"},
            "TABLE_NAME": "release_request",
        },
        "CONTACT_AUTHOR": {
            "ATTRIBUTES": {
                "DEP_SET_ID": "dep_set_id",
                "CITATION": "name_salutation",
                "NAME_FIRST": "name_first",
                "NAME_LAST": "name_last",
                "NAME_MI": "name_mi",
                "ROLE": "role",
                "EMAIL": "email",
                "ADDRESS_1": "address_1",
                "ADDRESS_2": "address_2",
                "ADDRESS_3": "address_3",
                "CITY": "city",
                "STATE_PROVINCE": "state_province",
                "POSTAL_CODE": "postal_code",
                "COUNTRY": "country",
                "PHONE": "phone",
                "FAX": "fax",
                "ORGANIZATION_TYPE": "organization_type",
            },
            "TABLE_NAME": "release_request",
        },
        "SITE": {"ATTRIBUTES": {"CODE": "code", "VERBOSE_NAME": "verbose_name"}, "TABLE_NAME": "site"},
        "DA_GROUP": {"ATTRIBUTES": {"CODE": "code", "GROUPNAME": "groupname", "SITE": "site", "MAIN_PAGE": "main_page"}, "TABLE_NAME": "da_group"},
        "SGCENTERS": {"ATTRIBUTES": {"CODE": "code", "VERBOSE_NAME": "verbose_name"}, "TABLE_NAME": "sgcenters"},
        "COMMUNICATION": {
            "ATTRIBUTES": {
                "ORDINAL": "ordinal",
                "SENDER": "sender",
                "RECEIVER": "receiver",
                "DEP_SET_ID": "dep_set_id",
                "WF_CLASS_ID": "wf_class_id",
                "WF_INST_ID": "wf_inst_id",
                "WF_CLASS_FILE": "wf_class_file",
                "COMMAND": "command",
                "STATUS": "status",
                "STATUS_TIMESTAMP": "status_timestamp",
                "PARENT_DEP_SET_ID": "parent_dep_set_id",
                "PARENT_WF_CLASS_ID": "parent_wf_class_id",
                "PARENT_WF_INST_ID": "parent_wf_inst_id",
            },
            "TABLE_NAME": "communication",
        },
        "ENGINE_MONITORING": {
            "ATTRIBUTES": {
                "CPU_ID": "cpu_id",
                "TOTAL_MEM": "total_mem",
                "TOTAL_PHYSICAL_MEM": "total_physical_mem",
                "TOTAL_VIRTUAL_MEM": "total_virtual_mem",
                "TOTAL_MEM_USAGE": "total_mem_usage",
                "PHYSICAL_MEM_USAGE": "physical_mem_usage",
                "VIRTUAL_MEM_USAGE": "virtual_mem_usage",
                "CPU_PROCESSES": "cpu_processes",
                "IDS_SET": "ids_set",
            },
            "TABLE_NAME": "engine_monitoring",
        },
    }

    # following lists are used in the WfDbApi

    _columnForStatus = ["STATUS_CODE", "INST_STATUS", "TASK_STATUS"]
    _objectTables = ["DEPOSITION", "WF_CLASS_DICT", "WF_INSTANCE", "WF_TASK", "WF_REFERENCE", "DA_USERS"]
    _tables = [
        "DEPOSITION",
        "WF_CLASS_DICT",
        "WF_INSTANCE",
        "WF_TASK",
        "WF_REFERENCE",
        "DA_USERS",
        "DATABASE_PDB_OBS_SPR",
        "DATABASE_REF",
        "DATABASE_RELATED",
        "AUTHOR_CORRECTIONS",
        "RELEASE_REQUEST",
        "DEP_WITH_PROBLEMS",
        "CONTACT_AUTHOR",
        "PROCESS_INFORMATION",
        "SITE",
        "DA_GROUP",
        "SGCENTERS",
        "COMMUNICATION",
        "ENGINE_MONITORING",
    ]
    _usefulItems = [
        "STATUS_CODE",
        "INST_STATUS",
        "TASK_STATUS",
        "DATABASE_CODE",
        "REPLACE_PDB_ID",
        "ID",
        "CONTENT_TYPE",
        "DB_ID",
        "DB_NAME",
        "RELATIONSHIP",
        "ASSOCIATED_IDS",
        "ASSESSION_CODE",
        "CORRECTIONS",
        "REQ_CITATION",
        "PROBLEM_TYPE",
        "PROBLEM_DETAILS",
        "PUBMED_ID",
    ]
    _objIds = ["DEP_SET_ID", "WF_CLASS_ID", "WF_INST_ID", "WF_TASK_ID"]
    _referencePairs = ["HASH_ID", "VALUE"]
    _userInfo = ["USER_NAME", "PASSWORD", "GROUPNAME", "EMAIL", "INITIALS"]

    _selectColumns = {
        2: [
            "deposition.dep_set_id",
            "deposition.pdb_id",
            "deposition.status_code",
            "deposition.author_release_status_code",
            "deposition.exp_method",
            "deposition.annotator_initials",
            "wf_class_dict.wf_class_id",
            "wf_class_dict.wf_class_name",
            "wf_class_dict.version",
            "wf_instance.wf_inst_id",
            "wf_instance.owner",
            "wf_instance.inst_status",
            "wf_instance.status_timestamp",
            "wf_task.wf_task_id",
            "wf_task.task_name",
            "wf_task.task_status",
            "wf_task.status_timestamp",
        ],
        1: [
            "DEP_SET_ID",
            "EXP_METHOD",
            "PDB_ID",
            "STATUS_CODE",
            "AUTHOR_RELEASE_STATUS_CODE",
            "INITIAL_DEPOSITION_DATE",
            "STATUS_CODE_EXP",
            "ANNOTATOR_INITIALS",
            "AUTHOR_LIST",
            "SG_CENTER",
        ],
    }
    _constraintList = {
        "DEP_SET_ID": "deposition.dep_set_id",
        "WF_CLASS_ID": "wf_class_dict.wf_class_id",
        "WF_INST_ID": "wf_instance.wf_inst_id",
        "WF_TASK_ID": "wf_task.wf_task_id",
        "STATUS_CODE": "deposition.status_code",
        "PDB_ID": "deposition.pdb_id",
        "DEPOSIT_SITE": "deposition.deposit_site",
        "PROCESS_SITE": "deposition.process_site",
        "ANNOTATOR_INITIALS": "deposition.annotator_initials",
        "EXP_METHOD": "deposition.exp_method",
        "SG_CENTER": "deposition.sg_center",
        "INST_STATUS": "wf_instance.inst_status",
        "INST_STATUS_TIMESTP": "wf_instance.status_timestamp",
        "TASK_STATUS_TIMESTP": "wf_task.status_timestamp",
        "TASK_STATUS": "wf_task.task_status",
        "OWNER": "wf_instance.owner",
    }

    # crossing search from wf_instance,wf_class_dict,deposition and wf_task

    _tableJoinSyntext = (
        " FROM wf_instance left join wf_class_dict on "
        + "wf_instance.wf_class_id= wf_class_dict.wf_class_id "
        + "left join deposition on (wf_instance.wf_class_id="
        + "wf_class_dict.wf_class_id and wf_instance.dep_set_id"
        + "=deposition.dep_set_id) left join wf_task on "
        + "(wf_task.wf_inst_id=wf_instance.wf_inst_id and "
        + "wf_task.dep_set_id=deposition.dep_set_id)"
    )

    _orderBy = {
        1: "order by dep_set_id",
        2: " order by deposition.dep_set_id, wf_instance.wf_inst_id desc," + "wf_instance.ordinal desc,wf_task.ordinal desc",
        3: ["author_release_status_code", "initial_deposition_date"],
    }

    def __init__(self):
        pass



##
# File:    WfDataObject.py
# Date:    5-April-2010
#
# Updates:
#    1-May-2015  jdw   add common output method
#    7-Sep-2015  jdw   add __str__ and __repl__
#
##
"""
Mixin container the workflow data object.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

try:
    import cStringIO as StringIO
except ImportError:
    import io as StringIO

from wwpdb.utils.wf.DataSelector import DataSelector
from wwpdb.io.locator.DataReference import DataFileReference
from wwpdb.utils.wf.DataValueContainer import DataValueContainer


class WfDataObject(DataSelector, DataValueContainer, DataFileReference):

    """Top-level container for workflow data object.

    This container includes some combination of the following:
    - A data file reference (`DataFileReference`)
    - A data value container (`DataValueContainer`)
    - A data selector (`DataSelector`)

    """

    # Inherit base __init__
    # def __init__(self):
    #    super(WfDataObject, self).__init__()
    # self.__verbose = False
    # sys.stderr.write("WfDataObject.__init_()\n")

    def printMe(self, ofh):
        if self.getReferenceType() in ["file"]:
            ofh.write("+WfDataObject.printMe() reference type    %s\n" % self.getReferenceType())
            ofh.write("+WfDataObject.printMe() version ID        %s\n" % self.getVersionId())
            ofh.write("+WfDataObject.printMe() data set ID       %s\n" % self.getDepositionDataSetId())
            ofh.write("+WfDataObject.printMe() partition number  %s\n" % self.getPartitionNumber())
            ofh.write("+WfDataObject.printMe() content type      %s\n" % self.getContentType())
            ofh.write("+WfDataObject.printMe() file format       %s\n" % self.getFileFormat())
            ofh.write("+WfDataObject.printMe() storage type      %s\n" % self.getStorageType())
            ofh.write("+WfDataObject.printMe() directory path    %s\n" % self.getDirPathReference())
            ofh.write("+WfDataObject.printMe() file path         %s\n" % self.getFilePathReference())
            if self.getSelectorType() is not None:
                ofh.write("+WfDataObject.printMe() selector type     %s\n" % self.getSelectorType())
                ofh.write("+WfDataObject.printMe() select category   %s\n" % self.getSelectCategoryName())
                ofh.write("+WfDataObject.printMe() select attributes %r\n" % self.getSelectAttributeList())
                ofh.write("+WfDataObject.printMe() select conditions %r\n" % self.getSelectConditionList())
        elif self.getContainerTypeName() in ["value", "list", "dict"]:
            ofh.write("+WfDataObject.printMe() container type %s\n" % self.getContainerTypeName())
            ofh.write("+WfDataObject.printMe() value type     %s\n" % self.getValueTypeName())
            ofh.write("+WfDataObject.printMe() is value set   %r\n" % self.isValueSet())
            ofh.write("+WfDataObject.printMe() is value valid %r\n" % self.isValueValid())
            ofh.write("+WfDataObject.printMe() data value     %r\n" % self.getValue())
        else:
            ofh.write("+WfDataObject.printMe() undefined data object\n")

    def __str__(self):
        output = StringIO.StringIO()
        self.printMe(ofh=output)
        contents = output.getvalue()
        output.close()
        return contents

    def __repr__(self):
        output = StringIO.StringIO()
        self.printMe(ofh=output)
        contents = output.getvalue()
        output.close()
        return contents


if __name__ == "__main__":
    wfd = WfDataObject()



##
# File:    WfTracking.py
# Date:    27-Apr-2010
#
# Updates:
#  07-Feb-2014   jdw   Move to common path and provide meaningful return code for setInstanceStatus()
#  14-Apr-2014   jdw   Update documentation  --
#
##
"""
Classes in this module manage the update of the progress and tracking
database for sequence editing operations.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import sys

from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi
from wwpdb.utils.wf.dbapi.WFEtime import getTimeNow


class WfTracking(object):
    """Provides methods to update progress and tracking information in the WF status database."""

    def __init__(self, verbose=False, log=sys.stderr):
        self.__verbose = verbose
        self.__lfh = log

    def setInstanceStatus(self, depId=None, instId=None, classId=None, status=None):
        """Update the tracking status record for the input instance step."""

        now = getTimeNow()
        DBstatusAPI = WfDbApi(verbose=self.__verbose)
        instD = {}
        instD["WF_INST_ID"] = instId
        instD["WF_CLASS_ID"] = classId
        instD["DEP_SET_ID"] = depId
        instD["INST_STATUS"] = status
        instD["STATUS_TIMESTAMP"] = now

        rd = DBstatusAPI.getObject(depId, classId, instId)
        if (rd is None) or (len(rd) == 0):
            self.__lfh.write("+WfTracking.setInstanceStatus() no status records for depId %s classId %s instId %s\n" % (depId, classId, instId))
            # new insert ---
            DBstatusAPI.saveObject(instD, "insert")
        #
        # This SQL is designed to avoid contention between modules.   No result (0 length) will be returned if
        #  module is currently running under the current workflow instance id.   If the workflow is running under
        # a different wf instance then we fail here --
        #
        #
        sql = (
            "update wf_instance_last set status_timestamp="
            + str(now)
            + ", inst_status='"
            + str(status)
            + "' where dep_set_id = '"
            + depId
            + "' and wf_class_id = '"
            + classId
            + "' and wf_inst_id = '"
            + instId
            + "'"
        )

        ok = DBstatusAPI.runUpdateSQL(sql)
        if ok < 1:
            self.__lfh.write("+WfTracking.setInstanceStatus() ERROR: failed to update workflow status, current task does not control the workflow\n")
            return False
        else:
            # Can update existing record using  ---
            DBstatusAPI.updateStatus(instD, status)
            # Verify the status
            if self.__verbose:
                rd = DBstatusAPI.getObject(depId, classId, instId)
                self.__lfh.write("+WfTracking.setInstanceStatus() verified new status is: %r\n" % DBstatusAPI.getStatus(rd))

        return True

    # Code broken and unused
    # def setTaskStatus(self,depId=None,instId=None,classId=None,taskId=None,status=None):
    #     """ Insert a tracking status record for a task step.
    #     """
    #     now = getTimeNow()
    #     taskD = {}
    #     taskD['WF_TASK_ID']   = taskId
    #     taskD['WF_INST_ID']   = instId
    #     taskD['WF_CLASS_ID']  = classId
    #     taskD['DEP_SET_ID']   = depId
    #     taskD['TASK_STATUS']  = status
    #     taskD['STATUS_TIMESTAMP'] = now

    #     rd = DBstatusAPI.getObject(depId,classId,instId,taskId)
    #     if ((rd is None) or (len(rd) == 0)):
    #         self.__lfh.write("+WfTracking.setTaskStatus() no status records for depId %s classId %s instId %s taskId %s\n" %
    #                          (depId,classId,instId,taskId))

    #     DBstatusAPI.saveObject(taskD, 'insert')

    #     # To update a current record ---
    #     #DBstatusAPI.updateStatus(taskID,"open")
    #     #DBstatusAPI.updateStatus(taskID,"closed(0)")
    #     #
    #     # Verify the status setting
    #     if (self.__verbose):
    #         rd = DBstatusAPI.getObject(depId,classId, instId,taskId)
    #         self.__lfh.write("+WfTracking.setTaskStatus() new status is: %r\n" % DBstatusAPI.getStatus(rd))

    #     return True
